import unittest

from srental.services import Contract, get_optimal_contracts, parse_contract_json


class TestContractService(unittest.TestCase):

    def test_empty_contract(self):
        # Given
        expected_result = []
        expected_profit = 0

        # When
        optimal_contracts, max_profit = get_optimal_contracts([])

        # Then
        self.assertEqual(optimal_contracts, expected_result)
        self.assertEqual(max_profit, expected_profit)

    def test_single_contract(self):
        # Given
        contract1 = Contract("Contract1", 0, 5, 10)
        expected_result = [contract1]
        expected_profit = 10

        # When
        optimal_contracts, max_profit = get_optimal_contracts([contract1])

        # Then
        self.assertEqual(optimal_contracts, expected_result)
        self.assertEqual(max_profit, expected_profit)

    def test_multiple_contract_with_different_durations(self):
        # Given
        contract1 = Contract("Contract1", 0, 5, 10)
        contract2 = Contract("Contract2", 7, 8, 14)
        contract3 = Contract("Contract3", 15, 6, 8)
        contracts = [contract1, contract2, contract3]
        expected_result = [contract1, contract2, contract3]
        expected_profit = 32

        # When
        optimal_contracts, max_profit = get_optimal_contracts(contracts)
        # Then
        self.assertEqual(optimal_contracts, expected_result)
        self.assertEqual(max_profit, expected_profit)

    def test_multiple_contract_with_larger_duration(self):
        # Given
        contract1 = Contract("Contract1", 15, 25, 30)
        contract2 = Contract("Contract2", 20, 26, 28)
        contracts = [contract1, contract2]
        expected_result = []
        expected_profit = 0

        # When
        optimal_contracts, max_profit = get_optimal_contracts(contracts)
        # Then
        self.assertEqual(optimal_contracts, expected_result)
        self.assertEqual(max_profit, expected_profit)

    def test_multiple_contract_with_negative_prices(self):
        # Given
        contract1 = Contract("Contract1", 2, 10, -5)
        contract2 = Contract("Contract2", 5, 6, -8)
        contracts = [contract1, contract2]
        expected_result = []
        expected_profit = 0

        # When
        optimal_contracts, max_profit = get_optimal_contracts(contracts)
        # Then
        self.assertEqual(optimal_contracts, expected_result)
        self.assertEqual(max_profit, expected_profit)

    def test_multiple_contract_with_overlaps(self):
        # Given
        contract1 = Contract("Contract1", 0, 5, 10)
        contract2 = Contract("Contract2", 3, 7, 14)
        contract3 = Contract("Contract3", 5, 9, 8)
        contract4 = Contract("Contract4", 5, 9, 7)

        expected_result = [contract1, contract3]
        expected_profit = 18

        # When
        contracts = [contract1, contract2, contract3, contract4]
        optimal_contracts, max_profit = get_optimal_contracts(contracts)

        # Then
        self.assertEqual(optimal_contracts, expected_result)
        self.assertEqual(max_profit, expected_profit)

    def test_multiple_contract_with_complex_combinations(self):
        # Given
        contract1 = Contract("Contract1", 0, 5, 10)
        contract2 = Contract("Contract2", 1, 6, 14)
        contract3 = Contract("Contract3", 5, 8, 10)
        contract4 = Contract("Contract4", 7, 9, 20)
        contract5 = Contract("Contract5", 18, 6, 30)
        contract6 = Contract("Contract5", 13, 4, 6)

        expected_result = [contract2, contract4, contract5]
        expected_profit = 64

        # When
        contracts = [contract1, contract2, contract3, contract4, contract5, contract6]
        optimal_contracts, max_profit = get_optimal_contracts(contracts)

        # Then
        self.assertEqual(optimal_contracts, expected_result)
        self.assertEqual(max_profit, expected_profit)


if __name__ == '__main__':
    unittest.main()



from typing import List


class Contract:
    def __init__(self, name: str, start: int, duration: int, price: int):
        self.name = name
        self.start = start
        self.duration = duration
        self.price = price

    def __repr__(self):
        return f"Contract(name={self.name}, start={self.start}, duration={self.duration}, price={self.price})"


def parse_contract_json(json_request):
    """
    Parse JSON data for a contract and return a Contract object
    """
    contracts = []
    try:
        for contract_dict in json_request:
            contract = Contract(
                name=contract_dict['name'],
                start=contract_dict['start'],
                duration=contract_dict['duration'],
                price=contract_dict['price']
            )
            contracts.append(contract)
    except Exception as e:
        print(f"Error converting JSON to Contracts: {str(e)}")
    return contracts


def get_optimal_contracts(contracts: List[Contract]) -> (List[Contract], int):
    """
    Given a list of Contract objects, return a list of contracts that maximize profitability
    """

    def calculate_profit(contracts, n, current_profit):
        nonlocal max_profit, result

        if n == len(contracts):
            if current_profit > max_profit:
                max_profit = current_profit
                result = current_solution.copy()
            return

        contract = contracts[n]

        # Skip contract if duration is greater than 24 or overlaps with existing contracts
        if contract.duration > 24 or not is_valid_slot(contract):
            calculate_profit(contracts, n + 1, current_profit)
        else:
            # Include the contract in the solution
            current_solution.append(contract)
            calculate_profit(contracts, n + 1, current_profit + contract.price)
            current_solution.pop()  # Backtrack

        # TODO: Exclude the contract from the solution
        calculate_profit(contracts, n + 1, current_profit)

    def is_valid_slot(contract):
        for c in current_solution:
            if contract.start < c.start + c.duration and contract.start + contract.duration > c.start:
                return False
        return True

    max_profit = 0
    result = []
    current_solution = []

    calculate_profit(contracts, 0, 0)

    return result, max_profit



import logging

from flask import jsonify, request
from . import app
from ..services.contract import parse_contract_json, get_optimal_contracts

logger = logging.getLogger(__name__)


@app.route('/spaceship/optimize', methods=['POST'])
def optimize_spaceship():
    request_data = request.get_json()
    logger.info("Received spaceship optimize request!")
    contracts = parse_contract_json(request_data)
    optimal_contracts, max_profit = get_optimal_contracts(contracts)
    optimal_contract_names = [contract.name for contract in optimal_contracts]
    return jsonify({'income': max_profit, 'path': optimal_contract_names})


# default route
@app.route('/', defaults={'path': ''})
@app.route('/<path:path>', methods=['GET', 'POST', 'PUT', 'DELETE'])
def catch_all(path):
    return 'Endpoint not found, Currently only support:/spaceship/optimize', 404



import logging

from flask import Flask

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s] %(message)s',
                    handlers=[logging.StreamHandler()])


def create_app(config_class='config.ProductionConfig'):
    app = Flask(__name__)
    app.config.from_object(config_class)

    from .controllers import spaceship
    app.register_blueprint(spaceship.app)

    return app



from setuptools import setup, find_packages

setup(
    name='srental',
    version='0.0.1',
    author='MingXing Xiao',
    author_email='xiaomixin@gmail.com',
    description='Used to optimize the profitability of a small rental company with a single spaceship to rent.',
    include_package_data=True,
    packages=find_packages(),
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    install_requires=[
        'flask==2.2.3',
        'python-dotenv==1.0.0',
    ],
)
